"""
    pyznap
    ~~~~~~~~~~~~~~

    pyznap - ZFS snapshot tool written in python.

    :copyright: (c) 2018 by Yannick Boetzel.
    :license: GPLv3, see LICENSE for more details.
"""


__version__ = '1.0.0'
